# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2011, 2degrees Limited <egoddard@tech.2degreesnetwork.com>.
# All Rights Reserved.
#
# This file is part of djeneralize <https://github.com/2degrees/djeneralize>,
# which is subject to the provisions of the BSD at
# <http://dev.2degreesnetwork.com/p/2degrees-license.html>. A copy of the
# license should accompany this distribution. THIS SOFTWARE IS PROVIDED "AS IS"
# AND ANY AND ALL EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST
# INFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
#
##############################################################################

from collections import defaultdict

from django.db.models.query import QuerySet

from djeneralize import PATH_SEPERATOR
from djeneralize.utils import find_next_path_down

__all__ = ['SpecializedQuerySet']


class SpecializedQuerySet(QuerySet):
    """
    A wrapper around QuerySet to ensure specialized models are returned.
    
    """
    
    def __init__(self, *args, **kwargs):
        """
        :param final_specialization: Whether the specializations returned are
            the most specialized specializations or whether the direct
            specializations are used
        :type final_specialization: :class:`bool`
        
        """
        
        final_specialization = kwargs.pop('final_specialization', True)
        
        super(SpecializedQuerySet, self).__init__(*args, **kwargs)
        self._final_specialization = final_specialization
    
    def iterator(self):
        """
        Override the iteration to ensure what's returned are Specialized Model
        instances.
        
        """
        
        # Get the resource ids and types together:
        specializations_by_id = self._clone().values_list(
            'specialization_type', 'id'
            )
        
        # Transform this into a dictionary of IDs by type:
        ids_by_specialization = defaultdict(list)
        
        # and keep track of the IDs which respect the ordering specified in the
        # queryset:
        specialization_ids = []
        
        for specialization, id in specializations_by_id:
            ids_by_specialization[specialization].append(id)
            specialization_ids.append(id) 
        
        specialized_model_instances = {}
        
        # Add the sub-class instances into a single look-up 
        for specialization, ids in ids_by_specialization.items():
            if not self._final_specialization:
                # Coerce the specialization to only be the direct child of the
                # general model (self.model):
                specialization = find_next_path_down(
                    self.model._meta.specialization, specialization,
                    PATH_SEPERATOR
                    )
            
            sub_queryset = self.model._meta.specializations[
                specialization
                ].objects.all()
                
            # Copy any deferred loading over to the new querysets:
            sub_queryset.query.deferred_loading = self.query.deferred_loading 
            
            sub_instances = sub_queryset.in_bulk(ids)
            
            specialized_model_instances.update(sub_instances)
        
        for resource_id in specialization_ids:
            yield specialized_model_instances[resource_id]
            
    def annotate(self, *args, **kawrgs):
        raise NotImplementedError(
            "%s does not support annotations as these cannot be reliably copied"
            " to the specialized instances" % self.__class__.__name__                    
            )
    
    def get(self, *args, **kwargs):
        """
        Override get to ensure a specialized model instance is returned.
        
        :return: A specialized model instance
        
        """
        
        if 'specialization_type' in kwargs:
            # if the specialization is explicitly specified, use this to work out
            # which sub-class of the general model we'll use:
            specialization = kwargs.pop('specialization_type')
        else:
            try:
                specialization = super(SpecializedQuerySet, self)\
                    .filter(*args, **kwargs).values_list(
                        'specialization_type', flat=True
                        )[0]
            except IndexError:
                raise self.model.DoesNotExist(
                    "%s matching query does not exist." %
                    self.model._meta.object_name
                    )
        
        if not self._final_specialization:
            # Coerce the specialization to only be the direct child of the
            # general model (self.model):
            specialization = find_next_path_down(
                self.model._meta.specialization, specialization, PATH_SEPERATOR
                )
        
        try:
            return self.model._meta.specializations[specialization]\
                                   .objects.get(*args, **kwargs)
        except KeyError:
            raise self.model.DoesNotExist("%s matching query does not exist." %
                                          self.model._meta.object_name)
    def direct(self):
        """
        Set the _final_specialization attribute on a clone of this queryset to
        ensure only directly descended specializations are considered.
        
        :return: The cloned queryset
        :rtype: :class:`SpecializedQuerySet`
        
        """
        
        clone = self._clone()
        clone._final_specialization = False
        return clone
    
    def final(self):
        """
        Set the _final_specialization attribute on a clone of this queryset to
        ensure only terminal specializations are considered.
        
        :return: The cloned queryset
        :rtype: :class:`SpecializedQuerySet`
        
        """
        
        clone = self._clone()
        clone._final_specialization = True
        return clone
            
    def _clone(self, klass=None, setup=False, **kwargs):
        """
        Customize the _clone method of QuerySet to ensure the value of
        _final_specialization is copied across to the clone correctly.
        
        :rtype: :class:`SpecializedQuerySet`
        
        """
        
        clone = super(SpecializedQuerySet, self)._clone(klass, setup, **kwargs)
        clone._final_specialization = self._final_specialization
        
        return clone


# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2011, 2degrees Limited <egoddard@tech.2degreesnetwork.com>.
# All Rights Reserved.
#
# This file is part of djeneralize <https://github.com/2degrees/djeneralize>,
# which is subject to the provisions of the BSD at
# <http://dev.2degreesnetwork.com/p/2degrees-license.html>. A copy of the
# license should accompany this distribution. THIS SOFTWARE IS PROVIDED "AS IS"
# AND ANY AND ALL EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST
# INFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""Utilities for djeneralize"""

__all__ = ['find_next_path_down']


def find_next_path_down(current_path, path_to_reduce, separator):
    """
    Manipulate ``path_to_reduce`` so that it only contains one more level of
    detail than ``current_path``.
    
    :param current_path: The path used to determine the current level
    :type current_path: :class:`basestring`
    :param path_to_reduce: The path to find the next level down
    :type path_to_reduce: :class:`basestring`
    :param separator: The string used to separate the parts of path
    :type separator: :class:`basestring`
    :return: The path one level deeper than that of ``current_path``
    :rtype: :class:`unicode`  
    
    """
    
    # Determine the current and next levels:
    current_level = current_path.count(separator)
    next_level = current_level + 1
    
    # Reduce the path to reduce down to just one more level deep than the
    # current path depth:
    return u'%s%s' % (
        separator.join(
            path_to_reduce.split(separator, next_level)[:next_level]
            ), separator
        )


##############################################################################
#
# Copyright (c) 2011, 2degrees Limited <egoddard@tech.2degreesnetwork.com>.
# All Rights Reserved.
#
# This file is part of djeneralize <https://github.com/2degrees/djeneralize>,
# which is subject to the provisions of the BSD at
# <http://dev.2degreesnetwork.com/p/2degrees-license.html>. A copy of the
# license should accompany this distribution. THIS SOFTWARE IS PROVIDED "AS IS"
# AND ANY AND ALL EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST
# INFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
#
##############################################################################

import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, "README")).read()
version = open(os.path.join(here, "VERSION")).readline().rstrip()

setup(name="djeneralize",
      version=version,
      description="Generalizations of specialized models for Django",
      long_description=README,
      classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Application Frameworks"
        ],
      keywords="django generalization specialization model",
      author="2degrees Limited",
      author_email="2degrees-floss@googlegroups.com",
      url="https://github.com/2degrees/djeneralize",
      license="BSD (http://dev.2degreesnetwork.com/p/2degrees-license.html)",
      packages=find_packages(),
      py_modules=[],
      zip_safe=False,
      tests_require = [
        "coverage",
        "fixture",
        "nose",
        ],
      install_requires=[
        "Django >= 1.1",
        ],
      extras_require = {
        'nose': ["nose >= 0.11"],
        },
      test_suite="nose.collector",
    )



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2011, 2degrees Limited <egoddard@tech.2degreesnetwork.com>.
# All Rights Reserved.
#
# This file is part of djeneralize <https://github.com/2degrees/djeneralize>,
# which is subject to the provisions of the BSD at
# <http://dev.2degreesnetwork.com/p/2degrees-license.html>. A copy of the
# license should accompany this distribution. THIS SOFTWARE IS PROVIDED "AS IS"
# AND ANY AND ALL EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST
# INFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
#
##############################################################################

from django.db.models.manager import Manager

from djeneralize.query import SpecializedQuerySet

__all__ = ['SpecializationManager']

class SpecializationManager(Manager):
    """
    Customized manager to ensure that any QuerySet that is used always returns
    specialized model instances rather than generalized model instances.
    
    The manager can either return *final_specializations*, i.e. the most
    specialized specialization, or the direct specialization of the general
    model.
    
    """
    
    def __init__(self):
        """
        :param final_specialization: Whether the specializations returned are
            the most specialized specializations or whether the direct
            specializations are used
        :type final_specialization: :class:`bool`
        
        """
        
        super(SpecializationManager, self).__init__()
        
    
    def get_query_set(self):
        """
        Instead of returning a QuerySet, use SpecializedQuerySet instead
        
        :return: A specialized queryset
        :rtype: :class:`SpecializedQuerySet`
        
        """

        return SpecializedQuerySet(self.model)
    
    def direct(self):
        """
        Set the _final_specialization attribute on a clone of the queryset to
        ensure only directly descended specializations are considered.
        
        :return: The cloned queryset
        :rtype: :class:`SpecializedQuerySet`
        
        """
        
        return self.get_query_set().direct()
    
    def final(self):
        """
        Set the _final_specialization attribute on a clone of the queryset to
        ensure only terminal specializations are considered.
        
        :return: The cloned queryset
        :rtype: :class:`SpecializedQuerySet`
        
        """
        
        return self.get_query_set().final()



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2011, 2degrees Limited <egoddard@tech.2degreesnetwork.com>.
# All Rights Reserved.
#
# This file is part of djeneralize <https://github.com/2degrees/djeneralize>,
# which is subject to the provisions of the BSD at
# <http://dev.2degreesnetwork.com/p/2degrees-license.html>. A copy of the
# license should accompany this distribution. THIS SOFTWARE IS PROVIDED "AS IS"
# AND ANY AND ALL EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST
# INFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
#
##############################################################################
import re

from django.db.models.base import ModelBase, Model
from django.db.models.fields import FieldDoesNotExist, TextField
from django.db.models import signals

from djeneralize import PATH_SEPERATOR
from djeneralize.manager import SpecializationManager
from djeneralize.utils import find_next_path_down

__all__ = ['BaseGeneralizationMeta', 'BaseGeneralizationModel']

SPECIALIZATION_RE = re.compile(r'^\w+$')
"""Allowed characters in the specializations declaration in class Meta"""


# { Metaclass:
    
class BaseGeneralizationMeta(ModelBase):
    """The metaclass for BaseGeneralizedModel"""
    
    def __new__(cls, name, bases, attrs):
        
        # Get the declared Meta inner class before the super-metaclass removes
        # it:
        meta = attrs.get('Meta')
        
        # We must remove the specialization declarations from the Meta inner
        # class since ModelBase will raise a TypeError is it encounters these:
        if meta:
            specialization = meta.__dict__.pop('specialization', None)
        else:
            specialization = None 
            
        new_model = super(BaseGeneralizationMeta, cls).__new__(
            cls, name, bases, attrs
            )
        
        if new_model._meta.abstract:
            # This is an abstract base-class and no specializations should be
            # declared on the inner class:
            if specialization is not None:
                # We need to ensure this is actually None and not just evaluates
                # to False as we enforce that it's not declared:
                raise TypeError(
                    "Abstract models should not have a specialization declared "
                    "on their inner Meta class"
                    )
        elif BaseGeneralizationModel in bases:
            # This must be a direct descendant from the BaseGeneralizationModel.
            # Prepare the look-up mapping of specializations which the sub-
            # classes will update:
            new_model._meta.specializations = {}
            new_model._meta.specialization = PATH_SEPERATOR
            
            if specialization is not None:
                # We need to ensure this is actually None and not just evaluates
                # to False as we enforce that it's not declared:
                raise TypeError(
                    "General models should not have a specialization declared "
                    "on their inner Meta class"
                    )
                
        else:
            if specialization is None:
                raise TypeError(
                    "Specialized models must declare specialization on their "
                    "inner Meta class"
                    )
            
            if not SPECIALIZATION_RE.match(specialization):
                raise ValueError("Specializations must be alphanumeric string")
            
            parent_class = new_model.__base__

            new_model._meta.specializations = {}
            new_model._generalized_parent = parent_class
            
            path_specialization = '%s%s%s' % (
                parent_class._meta.specialization, specialization,
                PATH_SEPERATOR
                )

            # Calculate the specialization as a path taking into account the
            # specialization of any ancestors:
            new_model._meta.specialization = path_specialization
            
            # Update the specializations mapping on the General model so that it
            # knows to use this class for that specialization:
            ancestor = getattr(new_model, '_generalized_parent', None)
            while ancestor:
                ancestor._meta.specializations[
                    new_model._meta.specialization
                    ] = new_model
                ancestor = getattr(ancestor, '_generalized_parent', None)
            
            parent_class._meta.specializations[path_specialization] = new_model
            
        # TODO: Reliably copy declared specializtion managers from super-classes
        # to this class. This will require an alteration to the _meta Option
        # class to hold information about concreate_specialized_managers and
        # abstract_specialized managers and quite a lot of overriding of methods
        # on the Manger class
        
        return new_model

#}

# { Base abstract model:

class BaseGeneralizationModel(Model):
    """Base model from which all Generalized and Specialized models inherit"""
    
    
    __metaclass__ = BaseGeneralizationMeta
    
    specialization_type = TextField(db_index=True)
    """Field to store the specialization"""
    
    def __init__(self, *args, **kwargs):
        """
        If specialization_type is not set in kwargs, add this is the most
        specialized model, set specialization_type to match the specialization
        declared in Meta
        
        """
                    
        super(BaseGeneralizationModel, self).__init__(*args, **kwargs)
        
        # If we have a final specialization, and a specialization_type is not
        # specified in kwargs, set it to the default for this model:
        if ('specialization_type' not in kwargs and
            not self._meta.specializations):
            self.set_default_specialization()
    
    class Meta:
        abstract = True
        
    def get_as_specialization(self, final_specialization=True):
        """
        Get the specialized model instance which corresponds to the general
        case.
        
        :param final_specialization: Whether the specialization returned is
            the most specialized specialization or whether the direct
            specialization is used
        :type final_specialization: :class:`bool`
        :return: The specialized model corresponding to the current model
        
        """
        
        path = self.specialization_type
        
        if not final_specialization:
            # We need to find the path which is only one-step down from the
            # current level of specialization.
            path = find_next_path_down(
                self._meta.specialization, path, PATH_SEPERATOR
                )
    
        return self._meta.specializations[path].objects.get(
            pk=self.pk
            )
    
    def set_default_specialization(self):
        """
        Set the ``specialization_type`` of this instance to match that specified
        by its specialization in the inner class Meta.
        
        """
        
        self.specialization_type = self._meta.specialization 

#}

# { Signal handler 

def ensure_specialization_manager(sender, **kwargs):
    """
    Ensure that a BaseGeneralizationModel subclass contains a default
    specialization manager and sets the ``_default_specialization_manager``
    attribute on the class.
    
    :param sender: The class which emitted the class prepared signal which needs
        to be checked for the specialization manager
    :type sender: :class:`django.db.models.base.Model`
    
    """
    
    cls = sender
    
    if not issubclass(cls, BaseGeneralizationModel):
        return
    
    if cls._meta.abstract:
        return
    if not getattr(cls, '_default_specialization_manager', None):
        # Create the default specialization manager, if needed.
        try:
            cls._meta.get_field('specializations')
            raise ValueError(
                "Model %s must specify a custom SpecializationManager, because "
                "it has a field named 'specializations'" % cls.__name__
                )
        except FieldDoesNotExist:
            pass
        cls.add_to_class('specializations', SpecializationManager())
        cls._base_specializations_manager = cls.specializations
    elif not getattr(cls, '_base_specializations_manager', None):
        default_specialization_mgr = \
            cls._default_specializations_manager.__class__
        if (default_specialization_mgr is SpecializationManager or
                getattr(default_specialization_mgr, "use_for_related_fields", False)):
            cls._base_specializations_manager = cls._default_specialization_manager
        else:
            # Default manager isn't a plain Manager class, or a suitable
            # replacement, so we walk up the base class hierarchy until we hit
            # something appropriate.
            for base_class in default_specialization_mgr.mro()[1:]:
                if (base_class is SpecializationManager or
                        getattr(base_class, "use_for_related_fields", False)):
                    cls.add_to_class('_base_specializations_manager', base_class())
                    return
            raise AssertionError(
                "Should never get here. Please report a bug, including your "
                "model and model manager setup."
                )

signals.class_prepared.connect(ensure_specialization_manager)

#}